@@ -1,22 +1,22 @@
-# -*- coding: utf-8 -*-
 """Implements JSON version of xonsh history backend."""
+import collections
+import collections.abc as cabc
 import os
 import sys
-import time
-import builtins
-import collections
 import threading
-import collections.abc as cabc
+import time
+
+from xonsh.built_ins import XSH
 
 try:
     import ujson as json
 except ImportError:
     import json  # type: ignore
 
-from xonsh.history.base import History
-import xonsh.tools as xt
 import xonsh.lazyjson as xlj
+import xonsh.tools as xt
 import xonsh.xoreutils.uptime as uptime
+from xonsh.history.base import History
 
 
 def _xhj_gc_commands_to_rmfiles(hsize, files):
@@ -34,7 +34,7 @@ def _xhj_gc_commands_to_rmfiles(hsize, files):
     """
     n = 0
     ncmds = 0
-    for _, fcmds, f, _ in reversed(files):
+    for _, fcmds, _, _ in reversed(files):
         # `files` comes in with empty files included (now), don't need special handling to gc them here.
 
         if ncmds + fcmds > hsize:
@@ -44,7 +44,7 @@ def _xhj_gc_commands_to_rmfiles(hsize, files):
 
     cmds_removed = 0
     files_removed = files[:-n]
-    for _, fcmds, f, _ in files_removed:
+    for _, fcmds, _, _ in files_removed:
         cmds_removed += fcmds
 
     return cmds_removed, files_removed
@@ -57,54 +57,77 @@ def _xhj_gc_files_to_rmfiles(hsize, files):
 
 
 def _xhj_gc_seconds_to_rmfiles(hsize, files):
-    """Return age removed (the age of the *oldest* file) and list of history files to remove to get under the age limit."""
+    """Return excess duration and list of history files to remove to get under the age limit."""
     now = time.time()
-    oldest_ts = files[0][0]
     n = 0
 
-    for ts, _, f, _ in files:
+    for ts, _, _, _ in files:
         if (now - ts) < hsize:
             break
         n += 1
 
-    return (now - oldest_ts) if n > 0 else 0, files[:n]
+    rmfiles = files[:n]
+    size_over = now - hsize - rmfiles[0][0] if n > 0 else 0
+    return size_over, rmfiles
 
 
 def _xhj_gc_bytes_to_rmfiles(hsize, files):
     """Return the history files to remove to get under the byte limit."""
     n = 0
     nbytes = 0
-    for _, _, f, fsize in reversed(files):
+    for _, _, _, fsize in reversed(files):
         if nbytes + fsize > hsize:
             break
         nbytes += fsize
         n += 1
     bytes_removed = 0
     files_removed = files[:-n]
-    for _, _, f, fsize in files_removed:
+    for _, _, _, fsize in files_removed:
         bytes_removed += fsize
 
     return bytes_removed, files_removed
 
 
+def _xhj_get_data_dir():
+    dir = xt.expanduser_abs_path(
+        os.path.join(XSH.env.get("XONSH_DATA_DIR"), "history_json")
+    )
+    if not os.path.exists(dir):
+        os.makedirs(dir)
+    return dir
+
+
 def _xhj_get_history_files(sort=True, newest_first=False):
     """Find and return the history files. Optionally sort files by
     modify time.
     """
-    data_dir = builtins.__xonsh__.env.get("XONSH_DATA_DIR")
-    data_dir = xt.expanduser_abs_path(data_dir)
-    try:
-        files = [
-            os.path.join(data_dir, f)
-            for f in os.listdir(data_dir)
-            if f.startswith("xonsh-") and f.endswith(".json")
-        ]
-    except OSError:
-        files = []
-        if builtins.__xonsh__.env.get("XONSH_DEBUG"):
-            xt.print_exception("Could not collect xonsh history files.")
+    data_dirs = [
+        _xhj_get_data_dir(),
+        XSH.env.get("XONSH_DATA_DIR"),  # backwards compatibility, remove in the future
+    ]
+
+    files = []
+    for data_dir in data_dirs:
+        data_dir = xt.expanduser_abs_path(data_dir)
+        try:
+            files += [
+                os.path.join(data_dir, f)
+                for f in os.listdir(data_dir)
+                if f.startswith("xonsh-") and f.endswith(".json")
+            ]
+        except OSError:
+            if XSH.env.get("XONSH_DEBUG"):
+                xt.print_exception(
+                    f"Could not collect xonsh history json files from {data_dir}"
+                )
     if sort:
         files.sort(key=lambda x: os.path.getmtime(x), reverse=newest_first)
+
+    custom_history_file = XSH.env.get("XONSH_HISTORY_FILE", None)
+    if custom_history_file:
+        custom_history_file = xt.expanduser_abs_path(custom_history_file)
+        if custom_history_file not in files:
+            files.insert(0, custom_history_file)
     return files
 
 
@@ -132,7 +155,7 @@ class JsonHistoryGC(threading.Thread):
     def run(self):
         while self.wait_for_shell:
             time.sleep(0.01)
-        env = builtins.__xonsh__.env  # pylint: disable=no-member
+        env = XSH.env  # pylint: disable=no-member
         xonsh_debug = env.get("XONSH_DEBUG", 0)
         if self.size is None:
             hsize, units = env.get("XONSH_HISTORY_SIZE")
@@ -141,10 +164,10 @@ class JsonHistoryGC(threading.Thread):
         files = self.files(only_unlocked=True)
         rmfiles_fn = self.gc_units_to_rmfiles.get(units)
         if rmfiles_fn is None:
-            raise ValueError("Units type {0!r} not understood".format(units))
+            raise ValueError(f"Units type {units!r} not understood")
 
         size_over, rm_files = rmfiles_fn(hsize, files)
-        hist = getattr(builtins.__xonsh__, "history", None)
+        hist = getattr(XSH, "history", None)
         if hist is not None:  # remember last gc pass history size
             hist.hist_size = size_over + hsize
             hist.hist_units = units
@@ -176,7 +199,7 @@ class JsonHistoryGC(threading.Thread):
         This is sorted by the last closed time. Returns a list of
         (file_size, timestamp, number of cmds, file name) tuples.
         """
-        env = getattr(getattr(builtins, "__xonsh__", None), "env", None)
+        env = XSH.env
         if env is None:
             return []
 
@@ -217,7 +240,7 @@ class JsonHistoryGC(threading.Thread):
                         end="",
                         file=sys.stderr,
                     )
-            except (IOError, OSError, ValueError):
+            except (OSError, ValueError):
                 continue
         files.sort()  # this sorts by elements of the tuple,
         # the first of which just happens to be file mod time.
@@ -232,7 +255,7 @@ class JsonHistoryFlusher(threading.Thread):
         self, filename, buffer, queue, cond, at_exit=False, skip=None, *args, **kwargs
     ):
         """Thread for flushing history."""
-        super(JsonHistoryFlusher, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
         self.filename = filename
         self.buffer = buffer
         self.queue = queue
@@ -258,7 +281,7 @@ class JsonHistoryFlusher(threading.Thread):
 
     def dump(self):
         """Write the cached history to external storage."""
-        opts = builtins.__xonsh__.env.get("HISTCONTROL")
+        opts = XSH.env.get("HISTCONTROL", "")
         last_inp = None
         cmds = []
         for cmd in self.buffer:
@@ -275,14 +298,16 @@ class JsonHistoryFlusher(threading.Thread):
 
             cmds.append(cmd)
             last_inp = cmd["inp"]
-        with open(self.filename, "r", newline="\n") as f:
+        with open(self.filename, newline="\n") as f:
             hist = xlj.LazyJSON(f).load()
         load_hist_len = len(hist["cmds"])
         hist["cmds"].extend(cmds)
         if self.at_exit:
-            hist["ts"][1] = time.time()  # apply end time
+            # todo: check why this is here.
+            if "ts" in hist:
+                hist["ts"][1] = time.time()  # apply end time
             hist["locked"] = False
-        if not builtins.__xonsh__.env.get("XONSH_STORE_STDOUT", False):
+        if not XSH.env.get("XONSH_STORE_STDOUT", False):
             [cmd.pop("out") for cmd in hist["cmds"][load_hist_len:] if "out" in cmd]
         with open(self.filename, "w", newline="\n") as f:
             xlj.ljdump(hist, f, sort_keys=True)
@@ -334,7 +359,7 @@ class JsonCommandField(cabc.Sequence):
         queue.append(self)
         with self.hist._cond:
             self.hist._cond.wait_for(self.i_am_at_the_front)
-            with open(self.hist.filename, "r", newline="\n") as f:
+            with open(self.hist.filename, newline="\n") as f:
                 lj = xlj.LazyJSON(f, reopen=False)
                 rtn = lj["cmds"][key].get(self.field, self.default)
                 if isinstance(rtn, xlj.LJNode):
@@ -353,7 +378,15 @@ class JsonHistory(History):
     JsonHistory implements an extra action: ``diff``
     """
 
-    def __init__(self, filename=None, sessionid=None, buffersize=100, gc=True, **meta):
+    def __init__(
+        self,
+        filename=None,
+        sessionid=None,
+        buffersize=100,
+        gc=True,
+        save_cwd=None,
+        **meta,
+    ):
         """Represents a xonsh session's history as an in-memory buffer that is
         periodically flushed to disk.
 
@@ -361,7 +394,7 @@ class JsonHistory(History):
         ----------
         filename : str, optional
             Location of history file, defaults to
-            ``$XONSH_DATA_DIR/xonsh-{sessionid}.json``.
+            ``$XONSH_DATA_DIR/history_json/xonsh-{sessionid}.json``.
         sessionid : int, uuid, str, optional
             Current session identifier, will generate a new sessionid if not
             set.
@@ -376,13 +409,22 @@ class JsonHistory(History):
         super().__init__(sessionid=sessionid, **meta)
         if filename is None:
             # pylint: disable=no-member
-            data_dir = builtins.__xonsh__.env.get("XONSH_DATA_DIR")
-            data_dir = os.path.expanduser(data_dir)
-            self.filename = os.path.join(
-                data_dir, "xonsh-{0}.json".format(self.sessionid)
-            )
+            data_dir = _xhj_get_data_dir()
+            self.filename = os.path.join(data_dir, f"xonsh-{self.sessionid}.json")
         else:
             self.filename = filename
+
+        if self.filename and not os.path.exists(os.path.expanduser(self.filename)):
+            meta["cmds"] = []
+            meta["sessionid"] = str(self.sessionid)
+            with open(self.filename, "w", newline="\n") as f:
+                xlj.ljdump(meta, f, sort_keys=True)
+
+            try:
+                os.chmod(self.filename, 0o600)
+            except Exception:  # pylint: disable=broad-except
+                pass
+
         self.buffer = []
         self.buffersize = buffersize
         self._queue = collections.deque()
@@ -391,16 +433,18 @@ class JsonHistory(History):
         self._skipped = 0
         self.last_cmd_out = None
         self.last_cmd_rtn = None
-        meta["cmds"] = []
-        meta["sessionid"] = str(self.sessionid)
-        with open(self.filename, "w", newline="\n") as f:
-            xlj.ljdump(meta, f, sort_keys=True)
         self.gc = JsonHistoryGC() if gc else None
         # command fields that are known
         self.tss = JsonCommandField("ts", self)
         self.inps = JsonCommandField("inp", self)
         self.outs = JsonCommandField("out", self)
         self.rtns = JsonCommandField("rtn", self)
+        self.cwds = JsonCommandField("cwd", self)
+        self.save_cwd = (
+            save_cwd
+            if save_cwd is not None
+            else XSH.env.get("XONSH_HISTORY_SAVE_CWD", True)
+        )
 
     def __len__(self):
         return self._len - self._skipped
@@ -426,7 +470,7 @@ class JsonHistory(History):
         if not self.remember_history:
             return
 
-        opts = builtins.__xonsh__.env.get("HISTCONTROL")
+        opts = XSH.env.get("HISTCONTROL", "")
         skipped_by_ignore_space = "ignorespace" in opts and cmd.get("spc")
         if skipped_by_ignore_space:
             return None
@@ -434,6 +478,9 @@ class JsonHistory(History):
         self.buffer.append(cmd)
         self._len += 1  # must come before flushing
 
+        if not self.save_cwd and "cwd" in cmd:
+            del cmd["cwd"]
+
         try:
             del cmd["spc"]
         except KeyError:
@@ -504,7 +551,7 @@ class JsonHistory(History):
                 commands = json_file.load()["cmds"]
             except (json.decoder.JSONDecodeError, ValueError):
                 # file is corrupted somehow
-                if builtins.__xonsh__.env.get("XONSH_DEBUG") > 0:
+                if XSH.env.get("XONSH_DEBUG") > 0:
                     msg = "xonsh history file {0!r} is not valid JSON"
                     print(msg.format(f), file=sys.stderr)
                 continue
@@ -523,7 +570,7 @@ class JsonHistory(History):
         data["length"] = len(self)
         data["buffersize"] = self.buffersize
         data["bufferlength"] = len(self.buffer)
-        envs = builtins.__xonsh__.env
+        envs = XSH.env
         data["gc options"] = envs.get("XONSH_HISTORY_SIZE")
         data["gc_last_size"] = f"{(self.hist_size, self.hist_units)}"
         return data
@@ -543,6 +590,7 @@ class JsonHistory(History):
         self.inps = JsonCommandField("inp", self)
         self.outs = JsonCommandField("out", self)
         self.rtns = JsonCommandField("rtn", self)
+        self.cwds = JsonCommandField("cwd", self)
         self._len = 0
         self._skipped = 0
 
